import time
from Bio import SeqIO


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def get_cigar_stats(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


def get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len):

    mismatch_set_all_cigar = set()
    mismatch_set_long_M_cigars = set()
    for each_cigar in r1_ref_cigar_set:
        aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_splitter(each_cigar))
        mismatch_set_all_cigar.add(mismatch_pct)
        if aligned_len >= min_M_len:
            mismatch_set_long_M_cigars.add(mismatch_pct)

    min_mismatch = 'NA'
    if len(mismatch_set_all_cigar) > 0:
        min_mismatch = min(mismatch_set_all_cigar)
        if len(mismatch_set_long_M_cigars) > 0:
            min_mismatch = min(mismatch_set_long_M_cigars)

    return min_mismatch


def check_both_ends_clipping(cigar_splitted):

    both_ends_clipping = False
    if len(cigar_splitted) >= 3:
        if (cigar_splitted[0][-1] in ['S', 's']) and (cigar_splitted[-1][-1] in ['S', 's']):
            both_ends_clipping = True

    return both_ends_clipping


def get_unlinked_mag_end_seq(ref_in, ref_in_end_seq, end_seq_len, ctg_ignore_region_dict_rd1):
    ctg_ignore_region_dict_rd2 = dict()

    # get ref seqs subset
    ref_subset_handle = open(ref_in_end_seq, 'w')
    for ref_seq in SeqIO.parse(ref_in, 'fasta'):

        ref_seq_id = ref_seq.id
        ref_seq_len = len(ref_seq.seq)

        if ref_seq_len < end_seq_len * 2:
            ref_subset_handle.write('>%s\n' % ref_seq_id)
            ref_subset_handle.write('%s\n' % ref_seq.seq)

            # add to ctg_ignore_region_dict_rd2
            if ref_seq_id in ctg_ignore_region_dict_rd1:
                ctg_ignore_region_dict_rd2[ref_seq_id] = ctg_ignore_region_dict_rd1[ref_seq_id]
        else:
            ref_seq_left_end_id = '%s_l' % ref_seq_id
            ref_seq_right_end_id = '%s_r' % ref_seq_id
            ref_seq_left_end = ref_seq.seq[:end_seq_len]
            ref_seq_right_end = ref_seq.seq[-end_seq_len:]

            # write out left end
            ref_subset_handle.write('>%s\n' % ref_seq_left_end_id)
            ref_subset_handle.write('%s\n' % ref_seq_left_end)

            # write out right end
            ref_subset_handle.write('>%s\n' % ref_seq_right_end_id)
            ref_subset_handle.write('%s\n' % ref_seq_right_end)

            # add to ctg_ignore_region_dict_rd2
            if ref_seq_id in ctg_ignore_region_dict_rd1:
                current_seq_to_ignore_ends = ctg_ignore_region_dict_rd1[ref_seq_id]
                for end_to_ignore in current_seq_to_ignore_ends:
                    if end_to_ignore == 'left_end':
                        ctg_ignore_region_dict_rd2[ref_seq_left_end_id] = {'left_end'}
                    if end_to_ignore == 'right_end':
                        ctg_ignore_region_dict_rd2[ref_seq_right_end_id] = {'right_end'}
    ref_subset_handle.close()

    return ctg_ignore_region_dict_rd2


def parse_rd2_sam_gnm_worker(arguments_list):

    rd1_unlinked_mags_sam_bowtie_reformat_sorted = arguments_list[0]
    free_living_ctg_ref_file                     = arguments_list[1]
    min_M_len_ctg                                = arguments_list[2]
    mismatch_cutoff                              = arguments_list[3]
    round_2_ctg_end_seq_len_dict                 = arguments_list[4]
    rd2_with_both_mates                          = arguments_list[5]
    ctg_ignore_region_dict_2rd                   = arguments_list[6]

    free_living_ctg_ref_file_handle = open(free_living_ctg_ref_file, 'w')
    current_read_base = ''
    current_read_base_r1_ctg_ref_dict_rd2 = dict()
    current_read_base_r2_ctg_ref_dict_rd2 = dict()
    with open(rd1_unlinked_mags_sam_bowtie_reformat_sorted) as rd1_unlinked_mags_sam_bowtie_reformat_sorted_opened:
        for each_line in rd1_unlinked_mags_sam_bowtie_reformat_sorted_opened:
            if not each_line.startswith('@'):
                each_line_split = each_line.strip().split('\t')
                cigar = each_line_split[5]
                read_id = each_line_split[0]
                read_id_base = '.'.join(read_id.split('.')[:-1])
                read_strand = read_id.split('.')[-1]
                ref_id = each_line_split[2]
                ref_pos = int(each_line_split[3])

                if current_read_base == '':
                    current_read_base = read_id_base

                    if cigar != '*':
                        if read_strand == '1':
                            current_read_base_r1_ctg_ref_dict_rd2[ref_id] = {ref_pos: cigar}
                        if read_strand == '2':
                            current_read_base_r2_ctg_ref_dict_rd2[ref_id] = {ref_pos: cigar}

                elif read_id_base == current_read_base:

                    if cigar != '*':
                        if read_strand == '1':
                            if ref_id not in current_read_base_r1_ctg_ref_dict_rd2:
                                current_read_base_r1_ctg_ref_dict_rd2[ref_id] = {ref_pos: cigar}
                            else:
                                current_read_base_r1_ctg_ref_dict_rd2[ref_id][ref_pos] = cigar
                        if read_strand == '2':
                            if ref_id not in current_read_base_r2_ctg_ref_dict_rd2:
                                current_read_base_r2_ctg_ref_dict_rd2[ref_id] = {ref_pos: cigar}
                            else:
                                current_read_base_r2_ctg_ref_dict_rd2[ref_id][ref_pos] = cigar
                else:
                    ################################### analysis previous read refs ####################################

                    ctg_refs_to_ignore_rd2 = set()

                    ########## get lowest mismatch for r1/r2 ctg refs ##########

                    # get r1_ref_cigar_set
                    r1_ref_cigar_set = set()
                    for each_pos_dict in current_read_base_r1_ctg_ref_dict_rd2.values():
                        each_pos_dict_values = {each_pos_dict[i] for i in each_pos_dict}
                        r1_ref_cigar_set.update(each_pos_dict_values)

                    # get r2_ref_cigar_set
                    r2_ref_cigar_set = set()
                    for each_pos_dict in current_read_base_r2_ctg_ref_dict_rd2.values():
                        each_pos_dict_values = {each_pos_dict[i] for i in each_pos_dict}
                        r2_ref_cigar_set.update(each_pos_dict_values)

                    r1_ref_min_mismatch = get_min_mismatch_from_cigar_list(r1_ref_cigar_set, min_M_len_ctg)
                    r2_ref_min_mismatch = get_min_mismatch_from_cigar_list(r2_ref_cigar_set, min_M_len_ctg)

                    ########## filter r1 ctg refs ##########
                    r1_ctg_refs_passed_qc = {}
                    for r1_ctg_ref_rd2 in current_read_base_r1_ctg_ref_dict_rd2:
                        r1_matched_pos_dict = current_read_base_r1_ctg_ref_dict_rd2[r1_ctg_ref_rd2]
                        if len(r1_matched_pos_dict) > 1:
                            ctg_refs_to_ignore_rd2.add(r1_ctg_ref_rd2)
                        else:
                            r1_ctg_ref_pos = list(r1_matched_pos_dict.keys())[0]
                            r1_ctg_ref_cigar = r1_matched_pos_dict[r1_ctg_ref_pos]
                            r1_ctg_ref_cigar_splitted = cigar_splitter(r1_ctg_ref_cigar)

                            # check both end clip
                            both_end_clp = check_both_ends_clipping(r1_ctg_ref_cigar_splitted)
                            if both_end_clp is True:
                                ctg_refs_to_ignore_rd2.add(r1_ctg_ref_rd2)
                            else:
                                # check mismatch
                                r1_aligned_len, r1_aligned_pct, r1_clipping_len, r1_clipping_pct, r1_mismatch_pct = get_cigar_stats(r1_ctg_ref_cigar_splitted)
                                if r1_ref_min_mismatch == 'NA':
                                    ctg_refs_to_ignore_rd2.add(r1_ctg_ref_rd2)
                                elif (r1_mismatch_pct > r1_ref_min_mismatch) or (r1_mismatch_pct > mismatch_cutoff):
                                    ctg_refs_to_ignore_rd2.add(r1_ctg_ref_rd2)
                                else:
                                    # check aligned length
                                    if r1_aligned_len < min_M_len_ctg:
                                        ctg_refs_to_ignore_rd2.add(r1_ctg_ref_rd2)
                                    else:
                                        # check if clp in the middle
                                        clip_in_middle = False
                                        if ('S' in r1_ctg_ref_cigar) or ('s' in r1_ctg_ref_cigar):
                                            clip_in_middle = True
                                            if (r1_ctg_ref_cigar_splitted[0][-1] in ['S', 's']) and (r1_ctg_ref_pos == 1):
                                                clip_in_middle = False
                                            if (r1_ctg_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                                if (r1_ctg_ref_pos + r1_aligned_len - 1) == round_2_ctg_end_seq_len_dict[r1_ctg_ref_rd2]:
                                                    clip_in_middle = False

                                        if clip_in_middle is True:
                                            ctg_refs_to_ignore_rd2.add(r1_ctg_ref_rd2)
                                        else:
                                            # check if matched to regions need to be ignored
                                            matched_to_r1_ref_ignored_region_rd2 = False

                                            if r1_ctg_ref_rd2 in ctg_ignore_region_dict_2rd:
                                                r1_ctg_ref_ends_to_ignore_rd2 = ctg_ignore_region_dict_2rd[r1_ctg_ref_rd2]
                                                for to_ignore_region_rd2 in r1_ctg_ref_ends_to_ignore_rd2:
                                                    if to_ignore_region_rd2 == 'left_end':
                                                        if r1_ctg_ref_pos <= 50:
                                                            matched_to_r1_ref_ignored_region_rd2 = True
                                                    if to_ignore_region_rd2 == 'right_end':
                                                        aln_len, aln_pct, clp_len, clp_pct, mis_pct = get_cigar_stats(cigar_splitter(r1_ctg_ref_cigar))
                                                        if (round_2_ctg_end_seq_len_dict[r1_ctg_ref_rd2] - r1_ctg_ref_pos - aln_len) <= 50:
                                                            matched_to_r1_ref_ignored_region_rd2 = True

                                            if matched_to_r1_ref_ignored_region_rd2 is False:
                                                r1_ctg_refs_passed_qc[r1_ctg_ref_rd2] = [r1_ctg_ref_cigar]
                                            else:
                                                ctg_refs_to_ignore_rd2.add(r1_ctg_ref_rd2)

                    ########## filter r2 ctg refs ##########
                    r2_ctg_refs_passed_qc = {}
                    for r2_ctg_ref_rd2 in current_read_base_r2_ctg_ref_dict_rd2:
                        r2_matched_pos_dict = current_read_base_r2_ctg_ref_dict_rd2[r2_ctg_ref_rd2]
                        if len(r2_matched_pos_dict) > 1:
                            ctg_refs_to_ignore_rd2.add(r2_ctg_ref_rd2)
                        else:
                            r2_ctg_ref_pos = list(r2_matched_pos_dict.keys())[0]
                            r2_ctg_ref_cigar = r2_matched_pos_dict[r2_ctg_ref_pos]
                            r2_ctg_ref_cigar_splitted = cigar_splitter(r2_ctg_ref_cigar)

                            # check both end clip
                            both_end_clp = check_both_ends_clipping(r2_ctg_ref_cigar_splitted)
                            if both_end_clp is True:
                                ctg_refs_to_ignore_rd2.add(r2_ctg_ref_rd2)
                            else:
                                # check mismatch
                                r2_aligned_len, r2_aligned_pct, r2_clipping_len, r2_clipping_pct, r2_mismatch_pct = get_cigar_stats(r2_ctg_ref_cigar_splitted)
                                if r2_ref_min_mismatch == 'NA':
                                    ctg_refs_to_ignore_rd2.add(r2_ctg_ref_rd2)
                                elif (r2_mismatch_pct > r2_ref_min_mismatch) or (r2_mismatch_pct > mismatch_cutoff):
                                    ctg_refs_to_ignore_rd2.add(r2_ctg_ref_rd2)
                                else:
                                    # check aligned length
                                    if r2_aligned_len < min_M_len_ctg:
                                        ctg_refs_to_ignore_rd2.add(r2_ctg_ref_rd2)
                                    else:
                                        # check if clp in the middle
                                        clip_in_middle = False
                                        if ('S' in r2_ctg_ref_cigar) or ('s' in r2_ctg_ref_cigar):
                                            clip_in_middle = True
                                            if (r2_ctg_ref_cigar_splitted[0][-1] in ['S', 's']) and (r2_ctg_ref_pos == 1):
                                                clip_in_middle = False
                                            if (r2_ctg_ref_cigar_splitted[-1][-1] in ['S', 's']):
                                                if (r2_ctg_ref_pos + r2_aligned_len - 1) == round_2_ctg_end_seq_len_dict[r2_ctg_ref_rd2]:
                                                    clip_in_middle = False

                                        if clip_in_middle is True:
                                            ctg_refs_to_ignore_rd2.add(r2_ctg_ref_rd2)
                                        else:
                                            # check if matched to regions need to be ignored
                                            matched_to_r2_ref_ignored_region_rd2 = False

                                            if r2_ctg_ref_rd2 in ctg_ignore_region_dict_2rd:
                                                r2_ctg_ref_ends_to_ignore_rd2 = ctg_ignore_region_dict_2rd[r2_ctg_ref_rd2]
                                                for to_ignore_region_rd2 in r2_ctg_ref_ends_to_ignore_rd2:
                                                    if to_ignore_region_rd2 == 'left_end':
                                                        if r2_ctg_ref_pos <= 50:
                                                            matched_to_r2_ref_ignored_region_rd2 = True
                                                    if to_ignore_region_rd2 == 'right_end':
                                                        aln_len, aln_pct, clp_len, clp_pct, mis_pct = get_cigar_stats(cigar_splitter(r2_ctg_ref_cigar))
                                                        if (round_2_ctg_end_seq_len_dict[r2_ctg_ref_rd2] - r2_ctg_ref_pos - aln_len) <= 50:
                                                            matched_to_r2_ref_ignored_region_rd2 = True

                                            if matched_to_r2_ref_ignored_region_rd2 is False:
                                                r2_ctg_refs_passed_qc[r2_ctg_ref_rd2] = [r2_ctg_ref_cigar]
                                            else:
                                                ctg_refs_to_ignore_rd2.add(r2_ctg_ref_rd2)

                    ####################################################################################################

                    r1_ctg_refs_rd2_no_ignored = {key: value for key, value in r1_ctg_refs_passed_qc.items() if key not in ctg_refs_to_ignore_rd2}
                    r2_ctg_refs_rd2_no_ignored = {key: value for key, value in r2_ctg_refs_passed_qc.items() if key not in ctg_refs_to_ignore_rd2}

                    r1_ctg_refs_rd2_no_ignored_str_list = {('%s__cigar__%s' % (key, value[0])) for key, value in r1_ctg_refs_passed_qc.items() if key not in ctg_refs_to_ignore_rd2}
                    r2_ctg_refs_rd2_no_ignored_str_list = {('%s__cigar__%s' % (key, value[0])) for key, value in r2_ctg_refs_passed_qc.items() if key not in ctg_refs_to_ignore_rd2}

                    # print(r1_ctg_refs_passed_qc)
                    # print(r1_ctg_refs_rd2_no_ignored)
                    # print(r2_ctg_refs_passed_qc)
                    # print(r2_ctg_refs_rd2_no_ignored)
                    # print()

                    # only r1 has no_ignored alignments
                    if (len(r1_ctg_refs_rd2_no_ignored) > 0) and (len(r2_ctg_refs_rd2_no_ignored) == 0):
                        print(r1_ctg_refs_passed_qc)
                        print(r1_ctg_refs_rd2_no_ignored)
                        print(r1_ctg_refs_rd2_no_ignored_str_list)
                        print('%s.2\t%s\n' % (current_read_base, ','.join(r1_ctg_refs_rd2_no_ignored)))
                        print('%s.2\t%s\n' % (current_read_base, ','.join(r1_ctg_refs_rd2_no_ignored_str_list)))



                        if rd2_with_both_mates is True:
                            free_living_ctg_ref_file_handle.write('%s\t%s\n' % (current_read_base, ','.join(r1_ctg_refs_rd2_no_ignored)))
                        else:
                            free_living_ctg_ref_file_handle.write('%s.2\t%s\n' % (current_read_base, ','.join(r1_ctg_refs_rd2_no_ignored)))

                    # only r2 has no_ignored alignments
                    if (len(r1_ctg_refs_rd2_no_ignored) == 0) and (len(r2_ctg_refs_rd2_no_ignored) > 0):
                        if rd2_with_both_mates is True:
                            free_living_ctg_ref_file_handle.write('%s\t%s\n' % (current_read_base, ','.join(r2_ctg_refs_rd2_no_ignored)))
                        else:
                            free_living_ctg_ref_file_handle.write('%s.1\t%s\n' % (current_read_base, ','.join(r2_ctg_refs_rd2_no_ignored)))


                    ########################################### reset values ###########################################

                    current_read_base = read_id_base
                    current_read_base_r1_ctg_ref_dict_rd2 = dict()
                    current_read_base_r2_ctg_ref_dict_rd2 = dict()

                    if cigar != '*':
                        if read_strand == '1':
                            current_read_base_r1_ctg_ref_dict_rd2[ref_id] = {ref_pos: cigar}
                        if read_strand == '2':
                            current_read_base_r1_ctg_ref_dict_rd2[ref_id] = {ref_pos: cigar}

    free_living_ctg_ref_file_handle.close()


def get_regions_to_ignore_from_barrnap_output(combined_barrnap_gff, ctg_len_dict):

    ctg_ignore_region_dict = {}

    for each_line in open(combined_barrnap_gff):
        if (not each_line.startswith('#')) and ('16S_rRNA' in each_line):
            each_line_split = each_line.strip().split('\t')
            ctg_id = each_line_split[0]
            ctg_len = ctg_len_dict[ctg_id]
            start_pos = int(each_line_split[3])
            end_pos = int(each_line_split[4])
            len_16s = end_pos - start_pos + 1
            left_gap = start_pos - 1
            right_gap = ctg_len - end_pos - 1

            if left_gap <= 50:
                if ctg_id not in ctg_ignore_region_dict:
                    ctg_ignore_region_dict[ctg_id] = {'left_end'}
                else:
                    ctg_ignore_region_dict[ctg_id].add('left_end')

            if right_gap <= 50:
                if ctg_id not in ctg_ignore_region_dict:
                    ctg_ignore_region_dict[ctg_id] = {'right_end'}
                else:
                    ctg_ignore_region_dict[ctg_id].add('right_end')

    return ctg_ignore_region_dict



step_2_wd = '/Users/songweizhi/Desktop/tunning_rd2_Oral'

combined_1st_round_unlinked_mag_end_seq         = '%s/file_in/round_1_unlinked_gnm_end_500bp.fa' % step_2_wd
pwd_splitted_gnm_sam_file                       = '/Users/songweizhi/Desktop/tunning_rd2_Oral/file_in/round_1_unlinked_gnm_sorted.sam'
pwd_splitted_gnm_sam_free_living_ctg_ref_file   = '/Users/songweizhi/Desktop/tunning_rd2_Oral/file_in/round_1_unlinked_gnm_sorted.txt'
combined_1st_round_unlinked_mags                = '%s/file_in/round_1_unlinked_gnm.fa' % step_2_wd
combined_barrnap_gff                            = '%s/file_in/combined_barrnap.gff' % step_2_wd
combined_input_gnms                             = '%s/file_in/3_Oral_refined_MAGs_combined.fa' % step_2_wd
min_M_len_ctg       = 60
mismatch_cutoff     = 2
rd2_with_both_mates = False
end_seq_len         = 500

ctg_len_dict = {}
for each_ctg_record in SeqIO.parse(combined_input_gnms, 'fasta'):
    ctg_len_dict[each_ctg_record.id] = len(each_ctg_record.seq)

round_2_ctg_end_seq_len_dict = {}
for each_ctg_end_record in SeqIO.parse(combined_1st_round_unlinked_mag_end_seq, 'fasta'):
    round_2_ctg_end_seq_len_dict[each_ctg_end_record.id] = len(each_ctg_end_record.seq)

ctg_ignore_region_dict = get_regions_to_ignore_from_barrnap_output(combined_barrnap_gff, ctg_len_dict)

ctg_ignore_region_dict_rd2 = get_unlinked_mag_end_seq(combined_1st_round_unlinked_mags, combined_1st_round_unlinked_mag_end_seq, end_seq_len, ctg_ignore_region_dict)

arguments_list = [pwd_splitted_gnm_sam_file, pwd_splitted_gnm_sam_free_living_ctg_ref_file,
                  min_M_len_ctg, mismatch_cutoff, round_2_ctg_end_seq_len_dict,
                  rd2_with_both_mates, ctg_ignore_region_dict_rd2]

parse_rd2_sam_gnm_worker(arguments_list)




import os


def sep_path_basename_ext(file_in):

    # separate path and file name
    file_path, file_name = os.path.split(file_in)
    if file_path == '':
        file_path = '..'

    # separate file basename and extension
    file_basename, file_extension = os.path.splitext(file_name)

    return file_path, file_basename, file_extension


def check_both_ends_clipping(cigar_splitted):

    both_ends_clipping = False
    if len(cigar_splitted) >= 3:
        if (cigar_splitted[0][-1] in ['S', 's']) and (cigar_splitted[-1][-1] in ['S', 's']):
            both_ends_clipping = True

    return both_ends_clipping


def cigar_splitter(cigar):

    # get the position of letters
    letter_pos_list = []
    n = 0
    for each_element in cigar:
        if (each_element.isalpha() is True) or (each_element == '='):
            letter_pos_list.append(n)
        n += 1

    # split cigar
    index = 0
    cigar_splitted = []
    while index <= len(letter_pos_list) - 1:
        if index == 0:
            cigar_splitted.append(cigar[:(letter_pos_list[index] + 1)])
        else:
            cigar_splitted.append(cigar[(letter_pos_list[index - 1] + 1):(letter_pos_list[index] + 1)])
        index += 1

    return cigar_splitted


def remove_both_ends_clp(sam_in, sam_out):
    sam_out_handle = open(sam_out, 'w')
    for each_read in open(sam_in):
        each_read_split = each_read.strip().split('\t')
        if each_read.startswith('@'):
            sam_out_handle.write(each_read)
        else:
            cigar = each_read_split[5]
            if cigar == '*':
                sam_out_handle.write(each_read)
            else:
                both_ends_clp = check_both_ends_clipping(cigar_splitter(cigar))
                if both_ends_clp is False:
                    sam_out_handle.write(each_read)
    sam_out_handle.close()


def get_cigar_stats(cigar_splitted):

    # aligned_len: M I X =
    # clipping_len: S
    # mismatch_len: X I D
    # mismatch_pct = mismatch_len / aligned_len
    # aligned_pct  = aligned_len  / (aligned_len + clipping_len)
    # clipping_pct = clipping_len / (aligned_len + clipping_len)

    aligned_len = 0
    clipping_len = 0
    mismatch_len = 0
    for each_part in cigar_splitted:
        each_part_len = int(each_part[:-1])
        each_part_cate = each_part[-1]

        # get aligned_len
        if each_part_cate in {'M', 'm', 'I', 'i', 'X', 'x', '='}:
            aligned_len += each_part_len

        # get clipping_len
        if each_part_cate in ['S', 's']:
            clipping_len += each_part_len

        # get mismatch_len
        if each_part_cate in {'I', 'i', 'X', 'x', 'D', 'd'}:
            mismatch_len += each_part_len

    aligned_pct  = float("{0:.2f}".format(aligned_len * 100 / (aligned_len + clipping_len)))
    clipping_pct = float("{0:.2f}".format(clipping_len * 100 / (aligned_len + clipping_len)))
    mismatch_pct = float("{0:.2f}".format(mismatch_len * 100 / (aligned_len)))

    return aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct


def remove_clp_in_middle(sam_in, sam_out):

    sam_out_handle = open(sam_out, 'w')

    marker_len_dict = {}
    for each_read in open(sam_in):
        each_read_split = each_read.strip().split('\t')
        if each_read.startswith('@'):
            sam_out_handle.write(each_read)

            marker_id = ''
            marker_len = 0
            for each_element in each_read_split:
                if each_element.startswith('SN:'):
                    marker_id = each_element[3:]
                if each_element.startswith('LN:'):
                    marker_len = int(each_element[3:])
            marker_len_dict[marker_id] = marker_len
        else:
            cigar = each_read_split[5]

            # check if clp in the middle
            if ('S' not in cigar) and ('s' not in cigar):
                sam_out_handle.write(each_read)
            else:
                ref_id = each_read_split[2]
                ref_pos = int(each_read_split[3])
                cigar_splitted = cigar_splitter(cigar)
                r1_aligned_len, r1_aligned_pct, r1_clipping_len, r1_clipping_pct, r1_mismatch_pct = get_cigar_stats(cigar_splitted)

                clip_in_middle = True
                if (cigar_splitted[0][-1] in ['S', 's']) and (ref_pos == 1):
                    clip_in_middle = False
                if (cigar_splitted[-1][-1] in ['S', 's']):
                    if (ref_pos + r1_aligned_len - 1) == marker_len_dict[ref_id]:
                        clip_in_middle = False

                if clip_in_middle is False:
                    sam_out_handle.write(each_read)

    sam_out_handle.close()


def keep_best_matches_in_sam_keep_short_M(sam_in, min_M_len, sam_out):

    # get read_to_cigar_dict
    read_to_cigar_dict = {}
    for each_line in open(sam_in):
        each_line_split = each_line.strip().split('\t')
        if not each_line.startswith('@'):
            read_id = each_line_split[0]
            cigar = each_line_split[5]
            if cigar != '*':
                both_ends_clp = check_both_ends_clipping(cigar_splitter(cigar))
                if both_ends_clp is False:
                    if read_id not in read_to_cigar_dict:
                        read_to_cigar_dict[read_id] = {cigar}
                    else:
                        read_to_cigar_dict[read_id].add(cigar)

    # get min_mismatch for each read
    read_min_mismatch_dict = {}
    for each_read in read_to_cigar_dict:
        read_mismatch_set_all_M = set()
        read_mismatch_set_long_M = set()
        for each_cigar in read_to_cigar_dict[each_read]:
            aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_splitter(each_cigar))
            read_mismatch_set_all_M.add(mismatch_pct)
            if aligned_len >= min_M_len:
                read_mismatch_set_long_M.add(mismatch_pct)
        read_min_mismatch = min(read_mismatch_set_all_M)
        if len(read_mismatch_set_long_M) > 0:
            read_min_mismatch = min(read_mismatch_set_long_M)
        read_min_mismatch_dict[each_read] = read_min_mismatch

    sam_file_best_match_handle = open(sam_out, 'w')
    for each_line in open(sam_in):
        if each_line.startswith('@'):
            sam_file_best_match_handle.write(each_line)
        else:
            each_line_split = each_line.strip().split('\t')
            cigar = each_line_split[5]
            if cigar == '*':
                sam_file_best_match_handle.write(each_line)
            else:
                read_id = each_line_split[0]
                cigar_split = cigar_splitter(cigar)
                both_ends_clp = check_both_ends_clipping(cigar_splitter(cigar))
                if both_ends_clp is False:
                    aligned_len, aligned_pct, clipping_len, clipping_pct, mismatch_pct = get_cigar_stats(cigar_split)
                    if mismatch_pct <= (read_min_mismatch_dict[read_id] * 1.5):
                        sam_file_best_match_handle.write(each_line)
    sam_file_best_match_handle.close()


def remove_high_mismatch(sam_in, mismatch_cutoff, sam_out):
    sam_out_handle = open(sam_out, 'w')
    for each_read in open(sam_in):
        each_read_split = each_read.strip().split('\t')
        if each_read.startswith('@'):
            sam_out_handle.write(each_read)
        else:
            cigar = each_read_split[5]
            if cigar == '*':
                sam_out_handle.write(each_read)
            else:
                cigar_splitted = cigar_splitter(cigar)
                r1_aligned_len, r1_aligned_pct, r1_clipping_len, r1_clipping_pct, r1_mismatch_pct = get_cigar_stats(cigar_splitted)
                if r1_mismatch_pct <= mismatch_cutoff:
                    sam_out_handle.write(each_read)
    sam_out_handle.close()



input_sam       = '5_261.sam'
#input_sam       = 'linked_to_hc_99.sam'
mismatch_cutoff = 2

sam_file_path, sam_file_basename, sam_file_extension = sep_path_basename_ext(input_sam)
sam_file_one_end_clp                                        = '%s/%s_one_end_clp.sam'                                               % (sam_file_path, sam_file_basename)
sam_file_one_end_clp_no_middle                              = '%s/%s_one_end_clp_no_middle.sam'                                     % (sam_file_path, sam_file_basename)
sam_file_one_end_clp_reformatted                            = '%s/%s_one_end_clp_no_middle_reformatted.sam'                         % (sam_file_path, sam_file_basename)
sam_file_one_end_clp_reformatted_log                        = '%s/%s_one_end_clp_no_middle_reformatted.log'                         % (sam_file_path, sam_file_basename)
sam_file_one_end_clp_reformatted_best_match                 = '%s/%s_one_end_clp_no_middle_reformatted_best_match.sam'              % (sam_file_path, sam_file_basename)
sam_file_one_end_clp_reformatted_best_match_low_mismatch    = '%s/%s_one_end_clp_no_middle_reformatted_best_match_low_mismatch.sam' % (sam_file_path, sam_file_basename)
sam_file_sorted                                             = '%s/%s_sorted.sam'                                                    % (sam_file_path, sam_file_basename)
coverage_file                                               = '%s/%s_cov.txt'                                                       % (sam_file_path, sam_file_basename)


# filter mapping
remove_both_ends_clp(input_sam, sam_file_one_end_clp)
remove_clp_in_middle(sam_file_one_end_clp, sam_file_one_end_clp_no_middle)
bbmap_reformat_cmd = 'reformat.sh in=%s out=%s sam=1.4 2> %s' % (sam_file_one_end_clp_no_middle, sam_file_one_end_clp_reformatted, sam_file_one_end_clp_reformatted_log)
os.system(bbmap_reformat_cmd)

keep_best_matches_in_sam_keep_short_M(sam_file_one_end_clp_reformatted, 35, sam_file_one_end_clp_reformatted_best_match)

remove_high_mismatch(sam_file_one_end_clp_reformatted_best_match, mismatch_cutoff, sam_file_one_end_clp_reformatted_best_match_low_mismatch)

# sort mapping
cmd_samtools_sort = 'samtools sort %s -o %s' % (sam_file_one_end_clp_reformatted_best_match_low_mismatch, sam_file_sorted)
os.system(cmd_samtools_sort)

# get mean depth
cmd_samtools_coverage = 'samtools --ff 4 coverage %s -o %s' % (sam_file_sorted, coverage_file)
os.system(cmd_samtools_coverage)



'''
#rname	startpos	endpos	numreads	covbases	coverage	meandepth	meanbaseq	meanmapq
cami_hc_SILVA138_id99_50_subsample_5_261	1	1560	4587	1560	100	286.179	40	255
cami_hc_SILVA138_id99_50_subsample_5_261	1	1560	4587	1560	100	286.179	40	255

cami_hc_SILVA138_id99_50_subsample_5_261	1	1560	4117	1560	100	255.337	40	0.958

cami_hc_SILVA138_id99_50_subsample_100_894	1	1573	4130	1573	100	245.599	40	0.926
cami_hc_SILVA138_id99_50_subsample_100_793	1	1573	3743	1573	100	228.688	40	0.944
cami_hc_SILVA138_id99_50_subsample_50_608	1	1573	4061	1573	100	248.905	40	0.944
cami_hc_SILVA138_id99_50_subsample_100_891	1	1573	3115	1572	99.9364	189.449	40	1.01
cami_hc_SILVA138_id99_50_subsample_100_1029	1	1569	4388	1569	100	270.631	40	0.948
cami_hc_SILVA138_id99_50_subsample_100_1027	1	1569	4168	1569	100	256.609	40	0.946
cami_hc_SILVA138_id99_50_subsample_5_263	1	1560	4336	1560	100	269.605	40	0.961
cami_hc_SILVA138_id99_50_subsample_50_660	1	1478	2903	1478	100	197.816	40	1.02




cami_hc_SILVA138_id99_50_subsample_5_261	1	1560	609	1553	99.5513	37.4974	40	0.969

cami_hc_SILVA138_id99_50_subsample_100_894	1	1573	532	1573	100	30.5683	40	0.914
cami_hc_SILVA138_id99_50_subsample_100_793	1	1573	494	1571	99.8729	29.417	40	0.947
cami_hc_SILVA138_id99_50_subsample_50_608	1	1573	592	1573	100	36.1087	40	0.938
cami_hc_SILVA138_id99_50_subsample_100_891	1	1573	408	1569	99.7457	25.9282	40	1.53
cami_hc_SILVA138_id99_50_subsample_100_1029	1	1569	665	1569	100	41.144	40	0.959
cami_hc_SILVA138_id99_50_subsample_100_1027	1	1569	582	1569	100	36.334	40	0.966
cami_hc_SILVA138_id99_50_subsample_5_263	1	1560	620	1560	100	39.7391	40	1
cami_hc_SILVA138_id99_50_subsample_50_660	1	1478	349	1478	100	24.6698	40	1.11


'''


import pandas as pd
import numpy as np
from webcolors import hex_to_rgb
import plotly.graph_objects as go # Import the graphical object
from plotly.offline import download_plotlyjs, init_notebook_mode, plot, iplot


node_label = ["A1", "A2", "B1", "B2","B3", "C1", "C2"]
node_index_dict = {'A1': 0, 'A2': 1, 'B1': 2, 'B2': 3, 'B3': 4, 'C1': 5, 'C2': 6}


source = ['A1','A1','A1','A2','A2','A2','B1','B2','B2','B3','B3']
target = ['B1','B2','B3','B1','B2','B3','C1','C1','C2','C1','C2']
values = [ 10, 5, 15, 5, 20, 45, 15, 20, 5, 30, 30 ]

source_index = [node_index_dict[x] for x in source]
target_index = [node_index_dict[x] for x in target]
# [0, 0, 0, 1, 1, 1, 2, 3, 3, 4, 4]
# [2, 3, 4, 2, 3, 4, 5, 5, 6, 5, 6]


fig = go.Figure(data=[go.Sankey(node = dict(label = node_label),
                    link = dict(source = source_index,
                                target = target_index,
                                value = values))])

# With this save the plots
plot(fig,image_filename='sankey_plot_1',image='png',image_width=1000,image_height=600)

# And shows the plot
fig.show()


import os

# file in
strain = 'CP'
wd                  = '/Users/songweizhi/Desktop/distance/%s' % strain
fake_bin            = '%s/bin_%s.fna'                       % (wd, strain)
ref_genome          = '%s/ref_%s.fna'                       % (wd, strain)
ref_genome_gff      = '%s/%s.gff'                           % (wd, strain)
ctg_cov_cutoff      = 90
iden_cutoff         = 99.5

# file out
output_table        = '%s/%s_distance.txt'                  % (wd, strain)


# tmp file
blast_op            = '%s/%s_bin_vs_ref.tab'               % (wd, strain)
blast_op_best_hit   = '%s/%s_bin_vs_ref_best_hit.tab'      % (wd, strain)


blastn_parameters = '-evalue 1e-5 -outfmt "6 qseqid sseqid pident length mismatch gapopen qstart qend sstart send evalue bitscore qlen slen" -task blastn'
blastn_cmd        = 'blastn -query %s -subject %s -out %s %s' % (fake_bin, ref_genome, blast_op, blastn_parameters)
os.system(blastn_cmd)

get_best_hit_cmd  = 'BioSAK BestHit -i %s -o %s' % (blast_op, blast_op_best_hit)
os.system(get_best_hit_cmd)


total_hit_num = 0
qualified_hit_num = 0
ref_len_dict = {}
recovered_ref_dict = {}
for each_hit in open(blast_op_best_hit):
    match_split = each_hit.strip().split('\t')
    query = match_split[0]
    subject = match_split[1]
    subject_genome = subject.split('_')[0]
    iden = float(match_split[2])
    align_len = int(match_split[3])
    sstart = int(match_split[8])
    send = int(match_split[9])
    query_len = int(match_split[12])
    subject_len = int(match_split[13])
    query_bin_name = '_'.join(query.split('_')[:-1])
    subject_bin_name = '_'.join(subject.split('_')[:-1])
    coverage_q = float(align_len) * 100 / float(query_len)
    coverage_s = float(align_len) * 100 / float(subject_len)
    total_hit_num += 1
    if (coverage_q >= ctg_cov_cutoff) and (iden >= iden_cutoff):
        qualified_hit_num += 1
        current_recovered_ref_region = '-'.join([str(i) for i in sorted([sstart, send])])
        if subject not in recovered_ref_dict:
            recovered_ref_dict[subject] = [current_recovered_ref_region]
        else:
            recovered_ref_dict[subject].append(current_recovered_ref_region)
        if subject not in ref_len_dict:
            ref_len_dict[subject] = subject_len

rna_16s_region_dict = {}
for each_line in open(ref_genome_gff):
    if not each_line.startswith('#'):
        if ';' in each_line:
            line_split = each_line.strip().split('\t')
            ctg_id = line_split[0]
            gene_product = line_split[-1].split('=')[-1]
            if gene_product == '16S ribosomal RNA':
                current_16s_region = '-'.join([str(i) for i in sorted([int(line_split[3]), int(line_split[4])])])
                if ctg_id not in rna_16s_region_dict:
                    rna_16s_region_dict[ctg_id] = [current_16s_region]
                else:
                    rna_16s_region_dict[ctg_id].append(current_16s_region)


output_table_handle = open(output_table, 'w')
for ref_seq in recovered_ref_dict:
    ref_seq_recovered = recovered_ref_dict[ref_seq]
    for ctg_region in ref_seq_recovered:
        output_table_handle.write('%s\tcontig\t%s\t%s\n' % (ref_seq, ctg_region.split('-')[0], ctg_region.split('-')[1]))
for ref_seq in rna_16s_region_dict:
    ref_16s_recovered = rna_16s_region_dict[ref_seq]
    for region_16s in ref_16s_recovered:
        output_table_handle.write('%s\trRNA16S\t%s\t%s\n' % (ref_seq, region_16s.split('-')[0], region_16s.split('-')[1]))
output_table_handle.close()

